/*
    SrvCheck - checks the availability of a network service on a remote
               host, and returns a status.

    Copyright (C) Mateusz Viste 2010,2011,2012. All rights reserved.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define pVer "0.20"
#define pDate "2010,2011,2012"
#define pUrl "http://www.viste-family.net/mateusz/software/srvcheck/"


#ifdef __MINGW32__
    #define WINVER 0x0501
    #include <winsock2.h>
    #include <ws2tcpip.h>
  #else
    #include <arpa/inet.h> /* sockaddr_in */
    #include <netdb.h>     /* getaddrinfo() */
#endif

#include <errno.h>     /* errno */
#include <stdint.h>    /* uint8_t */
#include <stdio.h>     /* printf() */
#include <stdlib.h>    /* atoi() */
#include <string.h>    /* strcmp() */
#include <sys/time.h>  /* gettimeofday() */
#include <unistd.h>    /* close() */

/* include all probing plugins */
#include "ftp.c"       /* SrvCheck_ftp()    */
#include "gopher.c"    /* SrvCheck_gopher() */
#include "http.c"      /* SrvCheck_http()   */
#include "nntp.c"      /* SrvCheck_nntp()   */
#include "pop3.c"      /* SrvCheck_pop3()   */
#include "smtp.c"      /* SrvCheck_smtp()   */
#include "ssh.c"       /* SrvCheck_ssh()    */


int main(int argc, char *argv[]) {
  char *SrvProt = NULL;
  int SrvPort = 0;
  uint8_t SrvProto;
  char *SrvHostname = NULL;
  uint8_t QuietFlag = 0;
  int x;   /* general purpose variable... */
  #define CheckResultErrStr_maxlen 1024
  float CheckResultTiming = -1;
  char CheckResultErrStr[CheckResultErrStr_maxlen];
  int sock;                        /* Socket descriptor */
  struct timeval timestart, timestop;
  char *param_http_url = "/";
  int gairesult;
  struct addrinfo *resolvedname, restemplate;
  #ifdef __MINGW32__
    WSADATA wsadata;
  #endif

  /* Parse parameters passed on the command line */
  for (x = 1; x < argc; x++) {
    if (strcasecmp(argv[x], "-prot") == 0) {
        x += 1;
        if (x == argc) {
          printf("The -prot parameter requires a complement!\n");
          return(1);
        }
        SrvProt = argv[x];
      } else if (strcasecmp(argv[x], "-port") == 0) {
        x += 1;
        if (x == argc) {
          printf("The -port parameter requires a complement!\n");
          return(1);
        }
        SrvPort = atoi(argv[x]);
        if ((SrvPort < 1) || (SrvPort > 65535)) {
          printf("The -port complement must be contained in the range 1..65535!\n");
          return(1);
        }
      } else if (strcasecmp(argv[x], "-host") == 0) {
        x += 1;
        if (x == argc) {
          printf("The -host parameter requires a complement!\n");
          return(1);
        }
        SrvHostname = argv[x];
      } else if (strcasecmp(argv[x], "-quiet") == 0) {
        QuietFlag = 1;
      } else if (strcasecmp(argv[x], "--help") == 0) {
        printf("\n");
        printf("SrvCheck v%s Copyright (C) Mateusz Viste %s\n", pVer, pDate);
        printf(" homepage: %s\n", pUrl);
        printf("\n");
        printf("SrvCheck is a ping-like program, which can check the availability of a server,\n");
        printf("using specific (application protocol) requests. It can also perform checks\n");
        printf("the raw TCP protocol (looking for a successul TCP handshake).\n");
        printf("\n");
        printf("Usage: srvcheck <parameters>\n");
        printf("\n");
        printf(" -host <host> hostname or IP address of the server to probe.\n");
        printf(" -prot <prot> specify the protocol to use for probing (HTTP, Gopher, SMTP,\n");
        printf("              POP3, NNTP, FTP, SSH, TCP).\n");
        printf(" -port <port> the TCP port to probe to (use default ports for protocols if not\n");
        printf("              specified). This is mandatory for the 'TCP' probing.\n");
        printf(" -quiet       do not output anything to the screen on successful check.\n");
        printf(" -license     display the program's license.\n");
        printf("\n");
        printf("Some protocols allow to use additional parameters. Here is the list.\n");
        printf("HTTP:\n");
        printf(" -http_url <url> checks a specific URL on the server (default is '/')\n");
        printf("\n");
        printf("SrvCheck will return 0 on success (that is, if the checked server\n");
        printf("is reachable and responds properly), and non-zero on error.\n");
        printf("\n");
        return(2);
      } else if (strcasecmp(argv[x], "-license") == 0) {
        printf("\n");
        printf("SrvCheck v%s Copyright (C) Mateusz Viste %s\n", pVer, pDate);
        printf(" homepage: %s\n", pUrl);
        printf("\n");
        printf("This program is free software: you can redistribute it and/or modify it under\n");
        printf("the terms of the GNU General Public License as published by the Free Software\n");
        printf("Foundation, either version 3 of the License, or (at your option) any later\n");
        printf("version.\n");
        printf("\n");
        printf("This program is distributed in the hope that it will be useful, but WITHOUT ANY\n");
        printf("WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A\n");
        printf("PARTICULAR PURPOSE. See the GNU General Public License for more details.\n");
        printf("\n");
        printf("You should have received a copy of the GNU General Public License along with\n");
        printf("this program. If not, see <http://www.gnu.org/licenses/>.\n");
        printf("\n");
        return(3);
      } else if (strcasecmp(argv[x], "-http_url") == 0) {
        x += 1;
        if (x == argc) {
          printf("The -http_url parameter requires a complement!\n");
          return(1);
        }
        param_http_url = argv[x];
      } else {
        printf("Unrecognized parameter: %s\n", argv[x]);
        return(1);
    }
  }

  /*   Here we go!   */
  if (SrvHostname == NULL) {
    printf("Error: You must specify a hostname. See 'srvcheck --help' for help.\n");
    return(1);
  }

  if (SrvProt == NULL) {
      printf("Error: You must specify a protocol. See 'srvcheck --help' for help.\n");
      return(1);
    } else if (strcasecmp(SrvProt, "HTTP") == 0) {
      SrvProto = 0x00;
      if (SrvPort == 0) SrvPort = 80;
    } else if (strcasecmp(SrvProt, "GOPHER") == 0) {
      SrvProto = 0x01;
      if (SrvPort == 0) SrvPort = 70;
    } else if (strcasecmp(SrvProt, "SMTP") == 0) {
      SrvProto = 0x02;
      if (SrvPort == 0) SrvPort = 25;
    } else if (strcasecmp(SrvProt, "POP3") == 0) {
      SrvProto = 0x03;
      if (SrvPort == 0) SrvPort = 110;
    } else if (strcasecmp(SrvProt, "NNTP") == 0) {
      SrvProto = 0x04;
      if (SrvPort == 0) SrvPort = 119;
    } else if (strcasecmp(SrvProt, "FTP") == 0) {
      SrvProto = 0x05;
      if (SrvPort == 0) SrvPort = 21;
    } else if (strcasecmp(SrvProt, "SSH") == 0) {
      SrvProto = 0x06;
      if (SrvPort == 0) SrvPort = 22;
    } else if (strcasecmp(SrvProt, "TCP") == 0) {
      SrvProto = 0x07;
      if (SrvPort == 0) {
        printf("Error: You have to specify a valid port for the TCP test.\n");
        return(1);
      }
    } else {
      printf("Error: Invalid protocol. See 'srvcheck --help' for help.\n");
      return(1);
  }

  CheckResultErrStr[0] = 0; /* Set the ErrStr to be empty */

  /* On Windows, we need to init WSAStartup */
  #ifdef __MINGW32__
    if (WSAStartup(MAKEWORD(1,1), &wsadata) == SOCKET_ERROR) {
      printf("Error creating socket.");
      return -1;
    }
  #endif

  /* Resolve the server name */
  memset (&restemplate, 0, sizeof(restemplate));
  restemplate.ai_family = PF_INET;
  gairesult = getaddrinfo(SrvHostname, NULL, &restemplate, &resolvedname);
  if (gairesult != 0) {
      /* I am not freeaddrinfo() resolvedname here, because the call failed */
      snprintf(CheckResultErrStr, CheckResultErrStr_maxlen, "Could not resolve [%s]\n", gai_strerror(gairesult));
    } else {
      /* Initiate a socket */
      if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
          freeaddrinfo(resolvedname); /* deallocate the addrinfo structure */
          snprintf(CheckResultErrStr, CheckResultErrStr_maxlen, "socket() failed [%s]\n", strerror(errno));
        } else {
          struct sockaddr_in sa;
          gettimeofday(&timestart, NULL);  /* Save current time */
          /* Prepare the sockaddr struct for connect() */
          memcpy(&sa,resolvedname->ai_addr,sizeof(sa));
          sa.sin_port = htons(SrvPort);
          freeaddrinfo(resolvedname); /* deallocate the addrinfo structure */
          /* Connect to the remote host */
          if (QuietFlag == 0) printf("Connecting to %s on port TCP/%d... ", SrvHostname, SrvPort);
          if (connect(sock, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
              snprintf(CheckResultErrStr, CheckResultErrStr_maxlen, "connect() failed [%s]\n", strerror(errno));
            } else {
              if (QuietFlag == 0) printf("ok\n");
              /* Perform the test */
              if (SrvProto == 0x00) { /* HTTP */
                  SrvCheck_http(sock, SrvHostname, SrvPort, param_http_url, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x01) { /* GOPHER */
                  SrvCheck_gopher(sock, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x02) { /* SMTP */
                  SrvCheck_smtp(sock, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x03) { /* POP3 */
                  SrvCheck_pop3(sock, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x04) { /* NNTP */
                  SrvCheck_nntp(sock, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x05) { /* FTP */
                  SrvCheck_ftp(sock, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x06) { /* SSH */
                  SrvCheck_ssh(sock, CheckResultErrStr, CheckResultErrStr_maxlen);
                } else if (SrvProto == 0x07) { /* TCP */
                  /* Here I do nothing, it's just an empty TCP handshake test */
                } else {
                  snprintf(CheckResultErrStr, CheckResultErrStr_maxlen, "Internal error! Unknown protocol!");
              }
          }
      }
      /* Shutdown and close the socket */
      shutdown(sock, 2);
      close(sock);
      gettimeofday(&timestop, NULL);  /* Save current time */
      /* Compute elapsed time */
      CheckResultTiming = ((timestop.tv_sec - timestart.tv_sec) * 1000000);
      if (timestop.tv_usec > timestart.tv_usec) {
          CheckResultTiming += (timestop.tv_usec - timestart.tv_usec);
        } else {
          CheckResultTiming -= (timestart.tv_usec - timestop.tv_usec);
      }
      CheckResultTiming /= 1000;
  }

  #ifdef __MINGW32__
    WSACleanup();
  #endif

  /* Print out the result */
  if (CheckResultErrStr[0] == 0) { /* Probing successfull (no error returned) */
      if (QuietFlag == 0) printf("The remote server is up and responded correctly (test completed in %0.2fms).\n", CheckResultTiming);
    } else {
      printf("Error: %s\n", CheckResultErrStr);
      return(2);
  }
  return(0);
}
